;;; -*- mode: scheme; coding: utf-8 -*-
;;;
;;; util/concurrent/actor.scm - Actor
;;;  
;;;   Copyright (c) 2017  Takashi Kato  <ktakashi@ymail.com>
;;;   
;;;   Redistribution and use in source and binary forms, with or without
;;;   modification, are permitted provided that the following conditions
;;;   are met:
;;;   
;;;   1. Redistributions of source code must retain the above copyright
;;;      notice, this list of conditions and the following disclaimer.
;;;  
;;;   2. Redistributions in binary form must reproduce the above copyright
;;;      notice, this list of conditions and the following disclaimer in the
;;;      documentation and/or other materials provided with the distribution.
;;;  
;;;   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;;;   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;;;   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
;;;   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
;;;   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
;;;   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
;;;   TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
;;;   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
;;;   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
;;;   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
;;;   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
;;;  

#!nounbound
(library (util concurrent actor)
    (export actor? (rename actor <actor>) make-actor
	    actor-send-message!
	    actor-receive-message!
	    actor-running?
	    actor-wait!
	    actor-terminate!
	    actor-state
	    actor-start!
	    actor-interrupt!
	    
	    make-shared-queue-channel-actor
	    make-shared-priority-queue-channel-actor

	    make-shared-queue-channel
	    make-shared-priority-queue-channel

	    *actor-thread-name-factory*
	    )
    (import (rnrs)
	    (sagittarius threads) ;; need thread-interrrupt!
	    (sagittarius)
	    (srfi :39 parameters)
	    (util concurrent shared-queue))

(define *actor-thread-name-factory* 
  (make-parameter (lambda () (gensym "actor-thread-"))))
;; base actor record
;; actor has 2 channels to receive/send messages
(define-record-type (actor make-actor actor?)
  (fields (mutable state) thread receiver sender)
  (protocol
   (lambda (p)
     (lambda (task make-receiver make-sender . maybe-start?)
       ;; default start (for backward compatibility...
       (define start? (or (null? maybe-start?) (car maybe-start?)))
       (let-values (((receiver/actor sender/client) (make-receiver))
		    ((receiver/client sender/actor) (make-sender)))
	 (let* ((t (make-thread
		    (lambda ()
		      (define actor (thread-specific (current-thread)))
		      (guard (e (else (actor-state-set! actor 'error)
				      (raise e)))
			(task receiver/client sender/client)
			(actor-state-set! actor 'finished)))
		    ((*actor-thread-name-factory*))))
		(actor (p 'created t receiver/actor sender/actor)))
	   (thread-specific-set! t actor)
	   (if start?
	       (actor-start! actor)
	       actor)))))))

(define (make-shared-queue-channel . opt)
  (define sq (apply make-shared-queue opt))
  (define (receiver . opt) (apply shared-queue-get! sq opt))
  (define (sender v . opt) (apply shared-queue-put! sq v opt))
  (values receiver sender))

(define (make-shared-priority-queue-channel compare . opt)
  (define sq (apply make-shared-priority-queue compare opt))
  (define (receiver . opt) (apply shared-priority-queue-get! sq opt))
  (define (sender v . opt) (apply shared-priority-queue-put! sq v opt))
  (values receiver sender))

(define (actor-start! actor)
  (when (eq? 'created (actor-state actor))
    (thread-start! (actor-thread actor))
    (actor-state-set! actor 'running))
  actor)

(define (actor-interrupt! actor)
  (and (actor-running? actor)
       (thread-interrupt! (actor-thread actor))))

(define (actor-running? actor) (eq? (actor-state actor) 'running))

(define (actor-send-message! actor message . opt)
  (apply (actor-sender actor) message opt))

(define (actor-receive-message! actor . opt)
  (apply (actor-receiver actor) opt))

;; need this?
(define (actor-wait! actor . timeout)
  (apply thread-join! (actor-thread actor) timeout))
(define (actor-terminate! actor)
  (thread-terminate! (actor-thread actor))
  (actor-state-set! actor 'terminated))

;; common actors
(define (make-shared-queue-channel-actor task . opt)
  (apply make-actor task make-shared-queue-channel make-shared-queue-channel
	 opt))

(define (make-shared-priority-queue-channel-actor task compare . opt)
  (define (make-channel) (make-shared-priority-queue-channel compare))
  (apply make-actor task make-channel make-channel opt))

)

